package RRPC;

import java.util.*;

// A Clause is a set of Literals, implicitly disjoined.

public class Clause implements java.io.Serializable
{
    static final long serialVersionUID = 8327718253681461304L;
    private Vector _literals;

    public Clause (){
	_literals = new Vector();
    }
    
    public Clause (Literal l){
	this();
	addLiteral(l);
    }

    // copy constructor
    public Clause(Clause c) {
	_literals =
	    (Vector)c._literals.clone();
    }

    public Clause (Vector ls){
	_literals = ls;
    }

    public boolean isEmpty(){
	return (size() == 0);
    }
    
    public Literal literalAt(int i){
	return (Literal) _literals.elementAt(i);
    }
    
    public int size(){
	return _literals.size();
    }


    public void addLiteral (Literal l){
	// First, test to be sure the literal isn't Equality on the same
	// terms 
	Proposition p = l.proposition();
	if (p instanceof CompoundProposition){
	    CompoundProposition cp = (CompoundProposition) p;
	    if (cp.getRelationSymbol().name().equals("Equals")
		&&
		cp.getTerms().getFirst().toString().equals
		(cp.getTerms().getRest().getFirst().toString())){
		if (l instanceof PositiveLiteral){
		    // Add True instead
		    l = new PositiveLiteral(new TrueProposition());
		} else
		    {
			// If it's ~Equals(x,x), then return (no need to add
			// false to a clause
			return;
		    }
	    }
	}

	// When we add a literal, test for subsumption  
	// Since we're disjoining, it's okay to leave out the *subsuming*
	// clause
	for (int i = 0; i < _literals.size(); i++){
	    if
		(((Literal) _literals.elementAt(i)).subsumes(l)){
		// Literal i subsumes l, so put l in its place
		_literals.setElementAt(l, i);
		return;
	    } else
		if (l.subsumes((Literal) _literals.elementAt(i))) {
		    // l subsumes literal i, so give up on l
		return;
		}
	}
	// Nobody subsumed anybody else, so just add the literal
	_literals.addElement(l);
    } 

    public void removeLiteralAt(int i) {
	_literals.removeElementAt(i);
    } 

    public void removeLiteral(Literal l){
	_literals.removeElement(l);
    }


    // take <s> and substitute in a copy of <this> Clause as necessary.
    public Clause substitute(Substitution s) {
	Clause c = new Clause();
	for (int i = 0; i < _literals.size(); i++) {
	    // Literal oldL = (Literal)_literals.get(i);
	    Literal oldL = literalAt(i);
	    Literal newL;
	    Proposition oldP = oldL.proposition();
	    Proposition newP = (Proposition)oldP.substitute(s);

	    if (oldL.isNegative())
		newL = new NegativeLiteral(newP);
	    else
		newL = new PositiveLiteral(newP);   
	    c.addLiteral(newL);
	}
	return c;
    } 


    // this method renames all of the Variables in a copy of <this> Clause.
    // very inefficient algorithm.
    public Clause renameVariables() {
	Clause c = new Clause(this); // make a copy of it first
	Hashtable oldVars = new Hashtable();
	Vector oldVarV = new Vector();
	
	// find out all the variables in this clause
	for (int i = 0; i < c._literals.size(); i++) {
	    Literal l = (Literal)c._literals.get(i);
	    Vector lv = l.proposition().obtainVariables();

	    for (int j = 0; j < lv.size(); j++) {
		Variable v = (Variable)lv.get(j);
		if (!oldVarV.contains(v))
		   
		    oldVarV.addElement(v);
	    }
	}

	// remap
	for (int i = 0; i < oldVarV.size(); i++) {
	    Variable currV = (Variable)oldVarV.get(i);
	    
	    Variable newV = currV.rename();
	    oldVars.put(currV, newV);
	}

	// go over all the variables again and replace them
	for (int i = 0; i < c._literals.size(); i++) {
	    Literal oldLiteral = (Literal)c._literals.get(i);
	    Proposition t = oldLiteral.proposition();
	    Set oldVarsSet = oldVars.keySet();
	    for (Iterator j = oldVarsSet.iterator(); j.hasNext(); ) {
		Variable oldVar = (Variable)j.next();
		t = (Proposition)t.substituteVariable
		    (oldVar, (Variable)oldVars.get(oldVar));
	    }

	    c._literals.remove(i);
	    if (oldLiteral.isNegative())
		c._literals.insertElementAt(new
NegativeLiteral(t), i);
	    else
		c._literals.insertElementAt(new
PositiveLiteral(t), i);
	}

	return c;
    } 

    // Don't side-effect this clause
    public Clause concatenate (Clause c2) {
	// Check to be sure that this isn't the clause containing only T
	if (this.toString().equals("T") && c2.size() != 0){
	    return c2;
	}

	Clause c1 = new Clause(this);
	for (int i = 0; i < c2.size(); i++){
	    c1.addLiteral((Literal) c2.literalAt(i));
	}
	return c1;
    } 
    
    // Find two literals that unify, and do so, simplifying the clause
    // Return null if there are no such literals
    // If there are multiple ways to factor, we just pick one arbitrarily
    // (this is okay, because the user can just factor again)

    public Clause factor() {
	int size = _literals.size();
	Clause c = new Clause(this);
	for (int i = 0; i < size; i++){
	    for (int j = i+1; j < size; j++){
		Substitution s = literalAt(i).unify(literalAt(j));
		if (s != null){		   
		    c.removeLiteralAt(i);
		    return c.substitute(s);
		}
	    }
	}
	// Couldn't factor
	return null;
    }

    // Return a vector of positive literals whose relation is Equals
    // and whose arity is 2
    public Vector extractEqualsLiterals(){
	Vector results = new Vector();
	int size = _literals.size();
	for (int i = 0; i < size; i++){
	    Literal l = literalAt(i);
	    if (l instanceof PositiveLiteral){
		Proposition p = l.proposition();
		if (p instanceof CompoundProposition) {
		   
		    CompoundProposition cp = (CompoundProposition) p;
		    if ((cp.getRelationSymbol().name().equals("Equals"))
			&&
			(cp.getTerms().length() == 2)) {
			results.addElement(l);
		   
		    }
		}
	    }
	}
	return results;
    }

    // Substitute newTerm in for oldTerm wherever it occurs
    // term1 might be compound or a constant (not necessarily a var)
    // Okay to check for object equality with oldTerm
   
    public void substituteTerm(Term newTerm, Term oldTerm){
	int size = _literals.size();
	for (int i = 0; i < size; i++){
	    
	    literalAt(i).substituteTerm(newTerm, oldTerm);
	}
    }

    // Substitute newTerm in for oldTerm, but copy as we go
    public Clause substituteTermAndCopy(Term newTerm, Term oldTerm){
	int size =_literals.size();
	Vector result = new Vector();
	for (int i = 0; i < size; i++){
	    result.add(literalAt(i).substituteTermAndCopy(newTerm,
							  oldTerm));
	}	
	return new Clause(result);
    }
    
    // Return a vector of terms contained in this clause that unify with t
    public Vector getPossibleMatches(Term t){
	int size = _literals.size();
	Vector result = new Vector();
	for (int i = 0; i < size; i++){
	    addAllNonDuplicates(result, literalAt(i).getPossibleMatches(t));
	}
	return result;
    }
    
    // This should be part of the Vector class
    // Add all the elements of v2 to v1 that are not duplicates
    private static void addAllNonDuplicates(Vector v1, Vector v2){
	int size2 = v2.size();
	for (int i = 0; i < size2; i++){
	    if(!v1.contains(v2.elementAt(i))){
		v1.add(v2.elementAt(i));
	    }
	}
	
    }

    public Clause deepCopy(){
	int size = _literals.size();
	Vector result = new Vector();
	for (int i = 0; i < size; i++){
	   
	    result.add(literalAt(i).deepCopy());
	}
	return new Clause(result);
    }
  
 
    public String toString(){
	String result;
	if (_literals.size() == 0)
	    result = "F";
	else {
	    result = ((Literal) _literals.elementAt(0)).toString();
	    for (int i = 1; i < _literals.size(); i++) {
		result = result + " v " + 
		    ((Literal) _literals.elementAt(i)).toString();
	    }
	}
	return result;
    } 

    
}
